#ifndef _FS_BOARD_H
#define _FS_BOARD_H

#include "freesweeper.h"
#include "skin.h"
#include "square.h"
#include "videoevent.h"
#include "timer.h"
#include "clickevents.h"
#include "luckpattern.h"
#include "translation.h"

#include "board-exceptions.h"
#include "board-layout.h"
#include "board-path-element.h"
#include "board-event.h"
#include "board-border.h"

#include "board-data.h"

#include "MersenneTwister/MersenneTwister.h"

#ifdef _WIN32
#include <windows.h>
#include <wincrypt.h>
#endif

class board : public QWidget
{
	Q_OBJECT

	QPainter painter;

	MTRand rg;

	translation* T;
	skin* sk;
	timer* Timer;
	
	square** squares;

	QList<video_event> video;
	QList<board_layout> boards;		// the stack of played boards
	
	QList<path_element> mouse_path;
	
	QList<luck_pattern*> lp_library;
public:
	QList<clone_event> clone_events;
	clickevents click_events;
public:
	int size;	// width*height
	int real_width; // if width<crit_width, this variable is equal to crit_width
	
	// what structures are valid
	bool has_layout;
	bool initialized;
	bool has_opening_borders;

	bool marathon;
	QList<board_description> marathon_description;
	int n_finished_boards;
	
	// mouse state
	bool left,right,middle;
	bool shift_left;
	bool onedotfive;	// is true iff the right mouse button is still pressed after flagging
	bool chorded;		// is true iff one of buttons is still pressed after a chord
	int cur_x;		// current square coordinates and mouse position
	int cur_y;
	int cur_prec_x,cur_prec_y;
	
	// scrollbar state
	bool scroll_x,scroll_y;
	int cur_scr_x,cur_scr_y;

	// video state
	int n_viewed_boards;
	int video_duration;
	int next_event;
	int arrow_x,arrow_y;
	bool events_built;
	int last_processed_event;

	// board properties and counters
	board_data D;
	
	// mode-related settings
	bool editing;
	bool restart_on_first_click;
	bool luck_allowed,custom_allowed,cheat_allowed;
	
	// general settings
	QString player;
	bool middle_chord;	// middle button behavior
	bool reveal_after_end;	// reveal the board after blast
	bool nf;		// force non-flagging
	bool flag_mines_after_win;
	bool show_wasted_flags_after_end;
	int clear_board_threshold;	// auto f2 after early blasts
	
	bool qmarks;	// this is here because we honor stupid traditions
	
	int lives,cur_lives;

	// limit settings
	int n_iter;	// number of iterations to meet specified 3bv/zini limits
	int lower_limit,upper_limit;	// 3bv limits
	int zn_lower_limit,zn_upper_limit;	// zini limits
	int hzn_lower_limit,hzn_upper_limit;	// human zini limits
	
	// luck settings
	bool luck_library,luck_solver;
	int l_mines,l_mines_required;
	int l_current_solution;
	int n_luck_iterations;
	
	// cheat settings
	bool superclick,superflag; 	// left clicks on opened squares as chords; right clicks as autoflags
	bool elmar;	// left press = left press + left click
	bool nono;	// movements of mouse while its buttons are pressed generate clicks
	bool autoflag;
	bool probabilities;
	bool show_board;	// show the board after the first click
	bool lawnmower;		// flag all mines after the first click
	bool force_seven,force_eight;	// the board must contain a seven or an eight
	int n_probabilities_iterations;
	
	// clone events settings
	bool clone_events_enabled;
	int ce_hesitation_time;
	int ce_iterations;
	
	// how to show video
	bool v_mouse_path,v_openings,v_show_board;
	
	// iteration state
	int probabilities_iter,luck_iter;
	
	// probabilities
	QList<border*> p_brd;
	border* c_brd;
	int pmines;
	
	// temporary things
	// ZiNi
	int zn_closed;
	// event list
	int last_event;
	int cur_isl;
	int last_opened_tm;
	bool last_square_is_op;
	bool game_started;
	bool op_hit,op_reacted,isl_changed,square_opened;
	
	// AVF stuff
	QString timestamp_begin;
	
#ifdef _WIN32
	HCRYPTPROV hprov;
	void generate_seed();
#endif
public:
	board(translation* T,skin* sk,timer* Timer);
	~board();

	virtual void mousePressEvent(QMouseEvent* e);
	virtual void mouseReleaseEvent(QMouseEvent* e);
	virtual void mouseMoveEvent(QMouseEvent* e);
	virtual void paintEvent(QPaintEvent* e);

	void draw_video_stuff();
	void draw_mouse_path();
	void draw_opening_borders();
	void draw_probabilities();
	
	void set_state(boardstate new_state);
	void check_win();
	
	// mouse event processors
	void left_press(int x, int y, int prec_x, int prec_y);
	void left_press_with_shift(int x, int y, int prec_x, int prec_y);
	void left_click(int x, int y, int prec_x, int prec_y, bool real_click=true);
	void right_press(int x, int y, int prec_x, int prec_y);
	void right_click(int x, int y, int prec_x, int prec_y);
	void middle_press(int x, int y, int prec_x, int prec_y);
	void middle_click(int x, int y, int prec_x, int prec_y);
	void mouse_move(int x, int y, int prec_x, int prec_y);
	
	bool is_inside(int x,int y);
	bool is_inside_board(int x,int y);

	// generating the board
	void first_click(int x,int y,int prec_x,int prec_y);
	void generate(int safe_x,int safe_y);
	void generate_fast(int safe_x,int safe_y);	// apply only 3bv limits
	void generate_with_zini(int safe_x,int safe_y,bool zini_l,bool hzini_l);	// apply 3bv and zini limits

	void random_board(int safe_x,int safe_y);
	void random_board(int* final,int* coord);

	// initializing the board properties
	void init_board();
	void init_board_basic();
	void init_board_full();
	
	void init_numbers();
	void init_openings();
	void init_islands();
	void init_opening_borders();
	void init_mines_counters();

	void process_opening(int op,int x,int y);
	void process_island(int isl,int x,int y);

	void reinit_properties();	// reloads the sizes of ops and isls assuming the squares are initialized
	
	void get_bbbv_fast();
	bool is_bbbv(int x,int y);
	
	void get_bbbv();

	// ZiNi functions (see board-zini.cpp)
	int get_zini(bool human);
	int zn_3bv_around(int x,int y);
	void zn_hit_openings(int& zini);
	void zn_reveal(int x,int y);
	void zn_open(int x,int y);
	void zn_click(int& zini,int x,int y);
	void zn_flag(int& zini,int x,int y);
	void zn_chord(int& zini,int x,int y);
	void zn_flag_around(int& zini,int x,int y);
	void zini_step(int& zini,bool human);
	
	// reset functions
	void new_game();
	void restart(bool video=false);
	void restart_from_blast();
	
	void clear_board();	// removes both the mine layout and the in-game state of squares
	void erase_board();	// removes the mine layout
	
	// the main board functions
	void open(int x,int y);
	void set_flag(int x,int y);
	void unset_flag(int x,int y);
	void chord(int x,int y,bool onedotfive=false);
	
	void show_opening(int x,int y);

	// procedures managing the counters 
	void check_openings();
	void reset_counters();
	
	// end functions
	void show_wasted_flags();
	
	// resizing
	void resize(int w, int h, int n);
	void change_n_mines(int n);
	void resize_if_needed(int w,int h,int n);
	
	void correct_board(int& w,int& h,int& n);
	void correct_size(int& w, int& h);
	void correct_n_mines(int& n);
	
	void init_squares();
	void init_borders();
	void delete_squares();

	// editing the board
	void put_mine(int x, int y);
	
	// the procedures changing the picture of a square (see board-squares.cpp)
	void show(int x,int y);
	void show_after_end(int x,int y);
	void flag(int x,int y);
	void unflag(int x,int y);
	void show_question(int x,int y);
	void show_mine(int x,int y);
	void show_blast(int x,int y);
	void show_wasted_flag(int x,int y);
	void show_misflag(int x,int y);
	void push(int x,int y);
	void pop(int x,int y);
	
	void repaint_square(int x,int y,QPixmap* pix);
	void repaint_arrow(int x,int y);
	
	void push_around(int x,int y);
	void pop_around(int x,int y);

	// procedures that process the neighborhoods
	int mines_around(int x,int y);
	int flags_around(int x,int y);
	int closed_sq_around(int x,int y);
	int opened_sq_around(int x,int y);
	int opened_sq_around_without_border(int x,int y);
	// related to solvers
	int known_mines_around(int x,int y);
	int known_safe_around(int x,int y);
	int unsolved_sq_around(int x,int y);
	int unknown_sq_around(int x,int y);
	
	// video (see board-video.cpp)
	void start_video();
	void end_video();
	
	void write_preflags();
	
	void write_mouse_event(int event,int x,int y);	// x and y are mouse coordinates
	void write_scroll_event(int event,int x,int y);	// x and y are scroll positions
	void write_board_event(int event,int x,int y);	// x and y are square coordinates
	void write_game_event(int event);
	
	void replace_fail_with_unblast();
	
	void save_rawvf(const QString& file, gamemode mode);
	void load_rawvf(const QString& file);
	 
	void save_avf(const QString& file,gamemode mode);
	void load_avf(const QString& file);
	
	void play_video();
	void stop_video(); 
	
	void save_board(const QString& file);
	void load_board(const QString& file);
	
	// mouse path
	void path_add_left_click(int x,int y);
	void path_add_right_click(int x,int y);
	void path_add_double_click(int x,int y);
	void path_add_mouse_move(int x,int y);
	
	// board stack functions (board-video.cpp)
	void clear_boards();
	void push_board();	// save the current board layout
	void pop_board();	// restore a board from stack to continue the video
	
	// luck (board-luck.cpp)
	void init_luck_library();
	bool try_luck_library_r(int id,int x,int y,luck_pattern_element mine_from,luck_pattern_element mine_to);
	bool try_luck_library(int id,int x,int y,luck_pattern_element mine_from,luck_pattern_element mine_to);
	
	bool l_check(int s,int x,int y);
	void l_solve_border(int sq);
	
	bool apply_luck_library(int x,int y);
	bool apply_luck_solver(int x,int y);
	void move_mine(int x_from,int y_from,int x_to,int y_to);

	// cheat functions
	void apply_autoflag();
	void flag_around(int x,int y);
	void flag_all();
	
	// probabilities (see board-probabilities.cpp)
	void get_probabilities();
	void p_reset_border();
	void p_find_isolated();
	int p_init_border();
	void p_set_mines(int x,int y);
	void p_set_safe(int x,int y);
	void p_solve_border(int sq);
	void p_get_ext_info(int cur, int proc, int sum, int cur_num);

	int remaining_mines();
	
	// click events
	void make_event_list();
	void el_add_event(int event);
	void el_next_click(int& last_click);
	void el_check_guessed_click(int x,int y);
	void el_check_guessed_flag(int x,int y);
	bool el_is_possible(int x,int y,int mine);
	bool has_solutions(int sq);
	void el_show_opening(int x,int y);
	void el_click(int x,int y);
	void el_chord(int x,int y,bool odf);
	void el_check_finished();	
	void el_pop_board();
	
	void process_border_c(int border,int x,int y);
	void process_border_o(int border,int x,int y);
	
	void print_board();
public slots:
	void scrolling_x(int x);
	void scrolling_y(int y);
	
	void process_video(int time);

	void toggle_mouse_path(bool on);
	void toggle_opening_borders(bool on);
	void toggle_reveal(bool on);
signals:
	void resized();
	
	void force_new_game();
	
	void changed_state(boardstate state);
	
	// smiley
	void press();
	void click();
	// red mines counter
	void flag();
	void unflag();
	// label
	void display_player(QString player);
	
	void changed_properties();
	
	void video_finished();
	void video_mode(int mode);
	
	friend int test();
};

#endif

